#include <clocale> // pour tolower
#include <cwctype> // pour towlower
#include <iostream>
#include <vector>
#include <filesystem> // nécessite -std=c++17
#include <sstream> // pour std::wstringstream

#ifdef _WINDOWS_PLATFORM_
#include "Win.h"  // définit struct subpaths
#else
#ifdef _QT_PLATFORM_
#include "Qt.h"  // définit struct subpaths
#include <fstream>
#include <string> // déjà inclus dans Win.h
#include <algorithm> // déjà inclus dans Win.h
#endif
#endif

bool isDirectory(std::wstring p)
{
    std::filesystem::path p1=p;
    return std::filesystem::is_directory(p1);
}

std::wstring configFolderPath(void); // définie dans paths.cpp
void deleteFile(const wstring& fileName); // définie dans Qt.cpp


// renvoie le contenu d'un fichier topfind adapté à WinCaml
std::wstring topfind(std::wstring findlib_parentpath)
{
    return std::wstring(LR"(#directory ")") + findlib_parentpath + std::wstring(LR"(/findlib"
#directory "+compiler-libs";;
(* For OCaml-4.00. This directory will be later removed from path *)

(* First test whether findlib_top is already loaded. If not, load it now.
 * The test works by executing the toplevel phrase "Topfind.reset" and
 * checking whether this causes an error.
 *)
let exec_test s =
let l = Lexing.from_string s in
let ph = !Toploop.parse_toplevel_phrase l in
let fmt = Format.make_formatter (fun _ _ _ -> ()) (fun _ -> ()) in
try
Toploop.execute_phrase false fmt ph
with
_ -> false
in
let is_native =
(* one of the few observable differences... *)
Gc.((get()).stack_limit) = 0 in
let suffix =
if is_native then "cmxs" else "cma" in
if not(exec_test "Topfind.reset;;") then (
let findlibpath = ")") + findlib_parentpath + std::wstring(LR"(/findlib" in
    Topdirs.dir_load Format.err_formatter (findlibpath ^ "/findlib." ^ suffix);
    Topdirs.dir_load Format.err_formatter (findlibpath ^ "/findlib_top." ^ suffix);
);
;;
#remove_directory "+compiler-libs";;

(* The following is always executed. It is harmless if findlib was already
 * initialized
 *)

let is_native =
  (* one of the few observable differences... *)
  Gc.((get()).stack_limit) = 0 in
let pred =
  if is_native then "native" else "byte" in
Topfind.add_predicates [ pred; "toploop" ];
Topfind.don't_load ["findlib"];
Topfind.announce();;)");
}

// renvoie le contenu d'un fichier findlib.conf adapté à WinCaml
std::wstring findlib_conf(std::wstring ocamllibpath, std::wstring findlib_parentpath, wstring sep)
{
    return LR"(destdir=")" + findlib_parentpath + LR"("
path=")" + ocamllibpath + sep + findlib_parentpath + LR"("
ocamlc="ocamlc.opt"
ocamlopt="ocamlopt.opt"
ocamldep="ocamldep.opt"
ocamldoc="ocamldoc.opt"
)";
}

//inscription des fichiers précédents dans le répertoire de configuration de WinCaml
void writeToFile(std::wstring toWrite, std::wstring path)
{
    string toBytes(const std::wstring & wstr); // pour la conversion en utf-8, défini dans Qt.cpp
    std::fstream fs(toBytes(path).c_str(), std::fstream::out); // chemin utf-8
    if (!fs.fail())
    {
        fs << toBytes(toWrite) << std::endl; // contenu utf-8
        fs.close();
    }
}

void writeFindlibConfig(struct subpaths sp) {
    std::wstring cfp = configFolderPath(); // chemin se terminant par un "/"
    if (sp.findlib != L"")
    {
        writeToFile(topfind(sp.findlib), cfp + L"topfind");
        writeToFile(findlib_conf(sp.ocamllib, sp.findlib, sp.ocwin ? L";" : L":"), cfp + L"findlib.conf");
    }
    else
    {
        deleteFile(cfp + L"topfind");
        deleteFile(cfp + L"findlib.conf");
    }
}

template <typename Type>
void remove_duplicate(std::vector<Type>& vec) {
    std::sort(vec.begin(), vec.end());
    vec.erase(unique(vec.begin(), vec.end()), vec.end());
}

std::wstring wslpath(std::wstring path)
{
    std::replace(path.begin(), path.end(), L'\\', L'/');
    if (path.substr(0, 15) == L"//wsl.localhost")
    {
        std::wstring subpath = path.substr(2 + path.substr(2).find(L'/'));
        size_t n = subpath.substr(1).find(L'/');
        return subpath.substr(n + 1);
    }
    if (path.substr(1, 2) == L":/") {
        std::wstring ret = (std::wstring(L"/mnt/") + wchar_t(std::towlower(path[0])) + path.substr(2));
        return ret;
    }
    return path;
}

void parcours_recursif(const std::filesystem::path& dir, struct subpaths& paths, std::vector<std::wstring>& stubs)
{
    for (const std::filesystem::directory_entry& file : std::filesystem::recursive_directory_iterator(dir, std::filesystem::directory_options::skip_permission_denied))
    {
        if (std::filesystem::is_regular_file(file)) {
            if (file.path().filename().wstring() == L"ld.conf") paths.ocamllib = file.path().parent_path().wstring();
            if (file.path().parent_path().filename().wstring() == L"findlib") paths.findlib = file.path().parent_path().parent_path().wstring();
            if (file.path().parent_path().filename().wstring() == L"stublibs") stubs.push_back(std::wstring(file.path().parent_path().wstring()));
        }
    }
}

//struct subpaths sp = {.ocaml=false,.camllight=false,.ocwin=false,.clwin=false,.ocamllib=L"",.stublibs=L"",.findlib=L""};
struct subpaths sp = {false,false,false,false,L"",L"",L""};

// renvoie les chemins de sous dossiers intéressants du dossier ocaml
struct subpaths dir_subpaths0(const std::filesystem::path& dir, bool camlnew=false)
{
    if (!camlnew) return sp;
    sp.ocaml = false;
    sp.camllight = false;
    sp.ocwin = false;
    sp.clwin =false;
    sp.ocamllib = L"";
    sp.stublibs = L"";
    sp.findlib = L"";

    struct subpaths paths = sp;

    std::vector<std::wstring> stubs;
    std::vector<std::filesystem::directory_entry> v;
    for (const std::filesystem::directory_entry& de : std::filesystem::directory_iterator(dir))
    {
        if (dir.wstring() + L"/bin" == de.path().wstring() || dir.wstring() + L"\\bin" == de.path().wstring()) {
            for (const std::filesystem::directory_entry& d : std::filesystem::directory_iterator(de)) {
                if (d.path().filename().wstring() == L"ocamlrun.exe") {
                    paths.ocaml = true; paths.ocwin = true;
                }
                if (d.path().filename().wstring() == L"ocamlrun") {
                    paths.ocaml = true; paths.ocwin = false;
                }
                if (d.path().filename().wstring() == L"camlrun.exe") {
                    paths.camllight = true; paths.clwin = true;
                }
                if (d.path().filename().wstring() == L"camlrun") {
                    paths.camllight = true; paths.clwin = false;
                }
            }
        }
        else if ((de.path().filename().wstring())[0] != L'.' && de.path().filename().wstring() != L"man") {
            v.push_back(de);
        }
    }
    if (paths.camllight) {
        paths.camllight = std::filesystem::exists(dir.wstring() + L"/lib/caml-light/" + (paths.clwin ? L"caml_all.exe" : L"caml_all"));
    }
    if (paths.ocaml) {
        for_each(v.begin(), v.end(),
                 [&paths, &stubs](std::filesystem::directory_entry de)
                 {
                     parcours_recursif(de, paths, stubs);
                 }
                 );
        if (paths.ocamllib == L"") paths.ocaml = false;
        if (stubs.size() > 0) remove_duplicate(stubs);

        std::wstringstream wss;
        if (paths.ocwin) {
            if (stubs.size() > 0) {
                wss << stubs[0];
                std::for_each(stubs.begin() + 1, stubs.end(), [&wss](const std::wstring ws) { wss << L";" << ws; });
                paths.stublibs = wss.str();
            }
        }
        else {
            if (stubs.size() > 0) {
                wss << wslpath(stubs[0]);
                std::for_each(stubs.begin() + 1, stubs.end(), [&wss](const std::wstring ws) { wss << L":" << wslpath(ws); });
                paths.stublibs = wss.str();
            }
            paths.ocamllib = wslpath(paths.ocamllib);
            paths.findlib = wslpath(paths.findlib);
        }
        std::replace(paths.ocamllib.begin(), paths.ocamllib.end(), L'\\', L'/');
        std::replace(paths.findlib.begin(), paths.findlib.end(), L'\\', L'/');
        std::replace(paths.stublibs.begin(), paths.stublibs.end(), L'\\', L'/');
        if (paths.findlib != L"")
        {
            std::wstring cfp = configFolderPath();
        //    void writeToFile(std::wstring toWrite, std::wstring path);
        //    std::wstring topfind(std::wstring findlib_parentpath);
        //    std::wstring findlib_conf(std::wstring ocamllibpath, std::wstring findlib_parentpath, wstring sep);
            writeToFile(topfind(paths.findlib), cfp + L"topfind");
            writeToFile(findlib_conf(paths.ocamllib, paths.findlib, paths.ocwin ? L";" : L":"), cfp + L"findlib.conf");
        }
    }
    sp = paths;
    return paths;
}

struct subpaths dir_subpaths(const wstring &p, bool camlnew=false)
{
    return dir_subpaths0(p,camlnew);
}
